import math
from typing import TypedDict

from lucid import register_model

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.types import _DeviceType

from lucid.models.objdet.util import (
    ROIAlign,
    MultiScaleROIAlign,
    FPN,
    apply_deltas,
    bbox_to_delta,
    nms,
    iou,
    clip_boxes,
)

from lucid.models.imgclf.resnet import resnet_50, resnet_101


__all__ = [
    "FasterRCNN",
    "faster_rcnn_resnet_50_fpn",
    "faster_rcnn_resnet_101_fpn",
]


class _AnchorGenerator(nn.Module):
    def __init__(
        self, sizes: tuple[int, ...], ratios: tuple[float, ...], stride: int
    ) -> None:
        super().__init__()
        self.stride = stride
        base_anchors = []
        for size in sizes:
            area = float(size * size)
            for r in ratios:
                w = math.sqrt(area / r)
                h = w * r
                base_anchors.append([-w / 2, -h / 2, w / 2, h / 2])

        self.base_anchors: nn.Buffer
        self.register_buffer("base_anchors", Tensor(base_anchors, dtype=lucid.Float32))
        self._cache: dict[tuple[int, int, _DeviceType], Tensor] = {}

    @property
    def num_anchors(self) -> int:
        return len(self.base_anchors)

    def grid_anchors(self, feat_h: int, feat_w: int, device: _DeviceType) -> Tensor:
        key = (feat_h, feat_w, device)
        if key not in self._cache:
            self._cache.clear()

        if key in self._cache:
            return self._cache[key]

        shift_x = (lucid.arange(feat_w, device=device) + 0.5) * self.stride
        shift_y = (lucid.arange(feat_h, device=device) + 0.5) * self.stride

        y, x = lucid.meshgrid(shift_y, shift_x, indexing="ij")
        shifts = lucid.stack((x.ravel(), y.ravel(), x.ravel(), y.ravel()), axis=1)

        anchors = self.base_anchors.to(device).reshape(1, self.num_anchors, 4)
        anchors = anchors + shifts.reshape(-1, 1, 4)
        anchors = anchors.reshape(-1, 4)

        self._cache[key] = anchors
        return anchors


class _RPNHead(nn.Module):
    def __init__(self, in_channels: int, num_anchors: int) -> None:
        super().__init__()
        self.conv = nn.Conv2d(in_channels, in_channels, kernel_size=3, padding=1)
        self.cls_logits = nn.Conv2d(in_channels, num_anchors, kernel_size=1)
        self.bbox_pred = nn.Conv2d(in_channels, num_anchors * 4, kernel_size=1)

    def forward(self, x: Tensor) -> tuple[Tensor, Tensor]:
        x = F.relu(self.conv(x))
        logits = self.cls_logits(x)
        deltas = self.bbox_pred(x)

        return logits, deltas


class _RPNLoss(TypedDict):
    rpn_cls_loss: Tensor
    rpn_reg_loss: Tensor
    proposals: list[Tensor]


class _RegionProposalNetwork(nn.Module):
    def __init__(
        self,
        in_channels: int,
        anchor_generator: _AnchorGenerator,
        pre_nms_top_n: int = 6000,
        post_nms_top_n: int = 1000,
        nms_thresh: float = 0.7,
        score_thresh: float = 0.0,
    ) -> None:
        super().__init__()
        self.head = _RPNHead(in_channels, anchor_generator.num_anchors)
        self.anchor_generator = anchor_generator

        self.pre_nms_top_n = pre_nms_top_n
        self.post_nms_top_n = post_nms_top_n
        self.nms_thresh = nms_thresh
        self.score_thresh = score_thresh

    def forward(self, feature: Tensor, image_shape: tuple[int, int]) -> list[Tensor]:
        B, _, H, W = feature.shape
        logits, deltas = self.head(feature)

        logits = logits.transpose((0, 2, 3, 1)).reshape(B, -1)
        deltas = deltas.transpose((0, 2, 3, 1)).reshape(B, -1, 4)

        anchors = self.anchor_generator.grid_anchors(H, W, feature.device)
        proposals: list[Tensor] = []

        for b in range(B):
            scores = F.sigmoid(logits[b])
            boxes = apply_deltas(anchors, deltas[b], add_one=1.0)
            boxes = clip_boxes(boxes, image_shape)

            keep = scores > self.score_thresh
            if lucid.sum(keep) == 0:
                proposals.append(lucid.empty(0, 4, device=feature.device))
                continue

            scores = scores[keep]
            boxes = boxes[keep]
            order = lucid.argsort(scores, descending=True)
            if self.pre_nms_top_n:
                order = order[: self.pre_nms_top_n]

            boxes = boxes[order]
            scores = scores[order]
            keep_idx = nms(boxes, scores, self.nms_thresh)
            if self.post_nms_top_n:
                keep_idx = keep_idx[: self.post_nms_top_n]

            proposals.append(boxes[keep_idx])

        return proposals

    def get_loss(
        self,
        feats: Tensor,
        targets: list[dict[str, Tensor]],
        image_shape: tuple[int, int],
    ) -> _RPNLoss:
        B, _, H, W = feats.shape
        logits, deltas = self.head(feats)
        logits = logits.transpose((0, 2, 3, 1)).reshape(B, -1)
        deltas = deltas.transpose((0, 2, 3, 1)).reshape(B, -1, 4)

        anchors = self.anchor_generator.grid_anchors(H, W, feats.device)

        rpn_cls_loss = lucid.zeros((), dtype=lucid.Float32)
        rpn_reg_loss = lucid.zeros((), dtype=lucid.Float32)
        proposals: list[Tensor] = []

        for b in range(B):
            gt = targets[b]["boxes"].to(feats.device)
            if gt.size == 0:
                labels = lucid.zeros(
                    anchors.shape[0], dtype=lucid.Int32, device=feats.device
                )
                reg_tgt = lucid.zeros_like(anchors)
            else:
                ious = iou(anchors, gt)
                max_iou = lucid.max(ious, axis=1)
                argmax = lucid.argsort(ious, axis=1, descending=True)[:, 0]

                labels = lucid.full(
                    (anchors.shape[0],), -1, dtype=lucid.Int32, device=feats.device
                )
                labels[max_iou < 0.3] = 0
                labels[max_iou >= 0.7] = 1

                for gi in range(gt.shape[0]):
                    best = lucid.argmax(ious[:, gi])
                    labels[best] = 1

                reg_tgt = lucid.zeros(anchors.shape[0], 4, device=feats.device)
                pos_mask = labels == 1
                if lucid.sum(pos_mask) > 0:
                    matched = gt[argmax[pos_mask]]
                    reg_tgt[pos_mask] = bbox_to_delta(anchors[pos_mask], matched)

            pos_idx = (labels == 1).nonzero().squeeze(axis=1)
            neg_idx = (labels == 0).nonzero().squeeze(axis=1)

            num_pos = min(pos_idx.shape[0], 128)
            num_neg = min(neg_idx.shape[0], 256 - num_pos)

            perm_pos = lucid.random.permutation(pos_idx.shape[0])[:num_pos]
            perm_neg = lucid.random.permutation(neg_idx.shape[0])[:num_neg]
            samp = lucid.concatenate([pos_idx[perm_pos], neg_idx[perm_neg]], axis=0)

            cls_scores = F.sigmoid(logits[b][samp])
            cls_targets = labels[samp].astype(lucid.Float32)
            rpn_cls_loss += F.binary_cross_entropy(
                cls_scores, cls_targets, reduction="mean"
            )

            if num_pos > 0:
                reg_pred = deltas[b][pos_idx[perm_pos]]
                reg_true = reg_tgt[pos_idx[perm_pos]]
                rpn_reg_loss += F.huber_loss(reg_pred, reg_true, reduction="mean")

            sc = F.sigmoid(logits[b])
            bx = apply_deltas(anchors, deltas[b], add_one=1.0)
            bx = clip_boxes(bx, image_shape)

            keep = sc > self.score_thresh
            if lucid.sum(keep) == 0:
                proposals.append(lucid.empty(0, 4, device=feats.device))
            else:
                sc_k, bx_k = sc[keep], bx[keep]
                order = lucid.argsort(sc_k, descending=True)[: self.pre_nms_top_n]
                bx_k, sc_k = bx_k[order], sc_k[order]

                keep_idx = nms(bx_k, sc_k, self.nms_thresh)[: self.post_nms_top_n]
                proposals.append(bx_k[keep_idx])

        rpn_cls_loss /= B
        rpn_reg_loss /= B

        return {
            "rpn_cls_loss": rpn_cls_loss,
            "rpn_reg_loss": rpn_reg_loss,
            "proposals": proposals,
        }


class _FasterRCNNLoss(TypedDict):
    rpn_cls_loss: Tensor
    rpn_reg_loss: Tensor
    roi_cls_loss: Tensor
    roi_reg_loss: Tensor
    total_loss: Tensor


class FasterRCNN(nn.Module):
    def __init__(
        self,
        backbone: nn.Module,
        feat_channels: int,
        num_classes: int,
        *,
        use_fpn: bool = False,
        anchor_sizes: tuple[int, ...] = (128, 256, 512),
        aspect_ratios: tuple[float, ...] = (0.5, 1.0, 2.0),
        anchor_stride: int = 16,
        pool_size: tuple[int, int] = (7, 7),
        hidden_dim: int = 4096,
        dropout: float = 0.5,
    ) -> None:
        super().__init__()
        self.backbone = backbone
        self.anchor_generator = _AnchorGenerator(
            anchor_sizes, aspect_ratios, anchor_stride
        )
        self.rpn = _RegionProposalNetwork(feat_channels, self.anchor_generator)
        self.use_fpn = use_fpn
        self.roipool = (
            MultiScaleROIAlign(pool_size) if self.use_fpn else ROIAlign(pool_size)
        )

        fc_in = feat_channels * pool_size[0] * pool_size[1]
        self.fc1 = nn.Linear(fc_in, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.drop1 = nn.Dropout(dropout)
        self.drop2 = nn.Dropout(dropout)

        self.cls_score = nn.Linear(hidden_dim, num_classes)
        self.bbox_pred = nn.Linear(hidden_dim, num_classes * 4)

        self.bbox_reg_means: nn.Buffer
        self.bbox_reg_stds: nn.Buffer
        self.register_buffer(
            "bbox_reg_means", Tensor([0.0, 0.0, 0.0, 0.0], dtype=lucid.Float32)
        )
        self.register_buffer(
            "bbox_reg_stds", Tensor([0.1, 0.1, 0.2, 0.2], dtype=lucid.Float32)
        )

    def roi_forward(
        self,
        feats: Tensor | list[Tensor],
        rois: Tensor,
        roi_idx: Tensor,
        *,
        return_feats: bool = False,
    ) -> tuple[Tensor, ...]:
        pooled = self.roipool(feats, rois, roi_idx)
        N = pooled.shape[0]

        x = pooled.reshape(N, -1)
        x = F.relu(self.fc1(x))
        x = self.drop1(x)
        x = F.relu(self.fc2(x))
        x = self.drop2(x)

        cls_logits = self.cls_score(x)
        bbox_deltas = self.bbox_pred(x)

        if return_feats:
            return cls_logits, bbox_deltas, feats
        return cls_logits, bbox_deltas

    def forward(
        self,
        images: Tensor,
        rois: Tensor | None = None,
        roi_idx: Tensor | None = None,
        *,
        return_feats: bool = False,
    ) -> tuple[Tensor, ...]:
        H, W = images.shape[2:]
        feats = self.backbone(images)
        if self.use_fpn and isinstance(feats, (list, tuple)):
            feature_rpn = feats[0]
            feature_roi = feats
        else:
            feature_rpn = feats
            feature_roi = feats

        if rois is None or roi_idx is None:
            proposals = self.rpn(feature_rpn, (H, W))
            proposals = [p.detach() for p in proposals]

            boxes_list, idx_list = [], []
            for i, p in enumerate(proposals):
                if p.size == 0:
                    continue
                boxes_list.append(p)
                idx_list.append(
                    lucid.full(
                        (p.shape[0],), i, dtype=lucid.Int32, device=images.device
                    )
                )
            if boxes_list:
                rois_px = lucid.concatenate(boxes_list, axis=0)
                roi_idx = lucid.concatenate(idx_list, axis=0)
            else:
                rois_px = lucid.empty(0, 4, device=images.device)
                roi_idx = lucid.empty(0, dtype=lucid.Int32, device=images.device)

            rois = rois_px / lucid.Tensor([W, H, W, H], dtype=lucid.Float32)
            rois = rois.to(images.device)
        else:
            rois_px = (rois * lucid.Tensor([W, H, W, H], dtype=lucid.Float32)).to(
                images.device
            )

        return self.roi_forward(feature_roi, rois, roi_idx, return_feats=return_feats)

    @lucid.no_grad()
    def predict(
        self,
        images: Tensor,
        *,
        score_thresh: float = 0.05,
        nms_thresh: float = 0.5,
        top_k: int = 100,
    ) -> list[dict[str, Tensor]]:
        B, _, H, W = images.shape
        feats = self.backbone(images)
        if self.use_fpn and isinstance(feats, (list, tuple)):
            feature_rpn = feats[0]
            feature_roi = feats
        else:
            feature_rpn = feats
            feature_roi = feats

        proposals = self.rpn(feature_rpn, (H, W))
        proposals = [p.detach() for p in proposals]

        boxes_list, idx_list = [], []
        for i, p in enumerate(proposals):
            if p.size == 0:
                continue

            boxes_list.append(p)
            idx_list.append(
                lucid.full((p.shape[0],), i, dtype=lucid.Int32, device=images.device)
            )

        if boxes_list:
            rois_px = lucid.concatenate(boxes_list, axis=0)
            roi_idx = lucid.concatenate(idx_list, axis=0)
        else:
            rois_px = lucid.empty(0, 4, device=images.device)
            roi_idx = lucid.empty(0, dtype=lucid.Int32, device=images.device)

        rois_norm = rois_px / lucid.Tensor([W, H, W, H], dtype=lucid.Float32)
        rois_norm = rois_norm.to(images.device)

        cls_logits, bbox_deltas = self.roi_forward(feature_roi, rois_norm, roi_idx)
        scores = F.softmax(cls_logits, axis=1)

        num_classes = scores.shape[1]
        detections = [{"boxes": [], "scores": [], "labels": []} for _ in range(B)]

        for c in range(1, num_classes):
            cls_scores = scores[:, c]
            deltas_cls = bbox_deltas[:, c * 4 : (c + 1) * 4]
            deltas_cls = deltas_cls * self.bbox_reg_stds + self.bbox_reg_means

            boxes_all = apply_deltas(rois_px, deltas_cls)
            boxes_all = clip_boxes(boxes_all, (H, W))

            mask = cls_scores > score_thresh
            if lucid.sum(mask) == 0:
                continue

            cls_scores = cls_scores[mask]
            boxes = boxes_all[mask]
            img_ids = roi_idx[mask]

            for img_id in img_ids.unique():
                m = img_ids == img_id
                boxes_i = boxes[m]
                scores_i = cls_scores[m]

                keep = nms(boxes_i, scores_i, nms_thresh)[:top_k]
                if keep.size == 0:
                    continue

                det = detections[int(img_id.item())]
                det["boxes"].append(boxes_i[keep])
                det["scores"].append(scores_i[keep])
                det["labels"].append(
                    lucid.full((keep.size,), c, dtype=lucid.Int32, device=images.device)
                )

        for det in detections:
            if det["boxes"]:
                det["boxes"] = lucid.concatenate(det["boxes"], axis=0)
                det["scores"] = lucid.concatenate(det["scores"], axis=0)
                det["labels"] = lucid.concatenate(det["labels"], axis=0)
            else:
                det["boxes"] = lucid.empty(0, 4, device=images.device)
                det["scores"] = lucid.empty(0, device=images.device)
                det["labels"] = lucid.empty(0, dtype=lucid.Int32, device=images.device)

        return detections

    def _match_rois(
        self, proposals: list[Tensor], targets: list[dict[str, Tensor]]
    ) -> tuple[Tensor, Tensor]:
        all_labels: list[Tensor] = []
        all_deltas: list[Tensor] = []

        for i, props in enumerate(proposals):
            gt_boxes = targets[i]["boxes"].to(props.device)
            gt_labels = targets[i]["labels"].to(props.device)

            if props.size == 0:
                all_labels.append(
                    lucid.empty(0, dtype=lucid.Int32, device=props.device)
                )
                all_deltas.append(lucid.empty(0, 4, device=props.device))
                continue

            ious = iou(props, gt_boxes)
            max_iou = lucid.max(ious, axis=1)
            argmax = lucid.argsort(ious, axis=1, descending=True)[:, 0]

            labels = gt_labels[argmax].astype(lucid.Int32)
            labels[max_iou < 0.5] = 0

            deltas = bbox_to_delta(props, gt_labels[argmax])
            all_labels.append(labels)
            all_deltas.append(deltas)

        roi_labels = lucid.concatenate(all_labels, axis=0)
        roi_reg_tgts = lucid.concatenate(all_deltas, axis=0)

        return roi_labels, roi_reg_tgts

    def get_loss(
        self, images: Tensor, targets: list[dict[str, Tensor]]
    ) -> _FasterRCNNLoss:
        B, _, H, W = images.shape
        feats = self.backbone(images)
        if self.use_fpn and isinstance(feats, (list, tuple)):
            feature_rpn = feats[0]
            feature_roi = feats
        else:
            feature_rpn = feats
            feature_roi = feats

        rpn_out = self.rpn.get_loss(feature_rpn, targets, (H, W))
        rpn_cls_loss = rpn_out["rpn_cls_loss"]
        rpn_reg_loss = rpn_out["rpn_reg_loss"]
        proposals = rpn_out["proposals"]
        proposals = [p.detach() for p in proposals]

        boxes_list, idx_list = [], []
        for i, props in enumerate(proposals):
            if props.size == 0:
                continue
            boxes_list.append(props)
            idx_list.append(
                lucid.full(
                    (props.shape[0],), i, dtype=lucid.Int32, device=images.device
                )
            )

        if boxes_list:
            rois_px = lucid.concatenate(boxes_list, axis=0)
            roi_idx = lucid.concatenate(idx_list, axis=0)
        else:
            rois_px = lucid.empty(0, 4, device=images.device)
            roi_idx = lucid.empty(0, dtype=lucid.Int32, device=images.device)

        rois_norm = rois_px / lucid.Tensor([W, H, W, H], dtype=lucid.Float32)
        rois_norm = rois_norm.to(images.device)

        sampled_rois, sampled_idx, sampled_labels, sampled_tgts = [], [], [], []
        roi_labels, roi_reg_targets = self._match_rois(proposals, targets)

        for i in range(B):
            mask_i = (roi_idx == i).nonzero().squeeze(axis=1)
            labels_i = roi_labels[mask_i]
            pos_i = (labels_i > 0).nonzero().squeeze(axis=1)
            neg_i = (labels_i == 0).nonzero().squeeze(axis=1)

            num_pos = min(pos_i.shape[0], 32)
            num_neg = min(neg_i.shape[0], 128 - num_pos)

            perm_pos = lucid.random.permutation(pos_i.shape[0])[:num_pos]
            perm_neg = lucid.random.permutation(neg_i.shape[0])[:num_neg]

            sel_idx = lucid.concatenate([pos_i[perm_pos], neg_i[perm_neg]], axis=0)
            sel = mask_i[sel_idx]

            sampled_rois.append(rois_norm[sel])
            sampled_idx.append(roi_idx[sel])
            sampled_labels.append(roi_labels[sel])
            sampled_tgts.append(roi_reg_targets[sel])

        sb_rois = lucid.concatenate(sampled_rois, axis=0)
        sb_idx = lucid.concatenate(sampled_idx, axis=0)
        sb_labels = lucid.concatenate(sampled_labels, axis=0)
        sb_tgts = lucid.concatenate(sampled_tgts, axis=0)

        cls_logits, bbox_deltas = self.roi_forward(feature_roi, sb_rois, sb_idx)
        roi_cls_loss = F.cross_entropy(cls_logits, sb_labels, reduction="mean")

        fg = (sb_labels > 0).nonzero().squeeze(axis=1)
        if fg.shape[0] > 0:
            preds = []
            for j, c in enumerate(sb_labels[fg]):
                start = c * 4
                preds.append(bbox_deltas[fg[j], start : start + 4])

            preds = lucid.stack(preds, axis=0)
            reg_tgt_fg = sb_tgts[fg]
            roi_reg_loss = F.huber_loss(preds, reg_tgt_fg, reduction="mean")

        else:
            roi_reg_loss = lucid.zeros((), dtype=lucid.Float32)

        total_loss = (rpn_cls_loss + rpn_reg_loss + roi_cls_loss + roi_reg_loss) / B
        return {
            "rpn_cls_loss": rpn_cls_loss,
            "rpn_reg_loss": rpn_reg_loss,
            "roi_cls_loss": roi_cls_loss,
            "roi_reg_loss": roi_reg_loss,
            "total_loss": total_loss,
        }


class _ResNetFPNBackbone(nn.Module):
    def __init__(self, net: nn.Module) -> None:
        super().__init__()
        self.net = net
        self.stem = self.net.stem
        self.maxpool = self.net.maxpool
        self.layers = [
            self.net.layer1,
            self.net.layer2,
            self.net.layer3,
            self.net.layer4,
        ]
        self.fpn = FPN([256, 512, 1024, 2048], out_channels=256)

    def forward(self, x: Tensor) -> Tensor:
        x = self.stem(x)
        x = self.maxpool(x)
        cs = []
        for i in range(4):
            x = self.layers[i](x)
            cs.append(x)
        return self.fpn(cs)


@register_model
def faster_rcnn_resnet_50_fpn(
    num_classes: int = 21, backbone_num_classes: int = 1000, **kwargs
) -> FasterRCNN:
    backbone = resnet_50(backbone_num_classes)
    backbone = _ResNetFPNBackbone(backbone)
    return FasterRCNN(
        backbone,
        feat_channels=256,
        num_classes=num_classes,
        use_fpn=True,
        hidden_dim=1024,
        **kwargs,
    )


@register_model
def faster_rcnn_resnet_101_fpn(
    num_classes: int = 21, backbone_num_classes: int = 1000, **kwargs
) -> FasterRCNN:
    backbone = resnet_101(backbone_num_classes)
    backbone = _ResNetFPNBackbone(backbone)
    return FasterRCNN(
        backbone,
        feat_channels=256,
        num_classes=num_classes,
        use_fpn=True,
        hidden_dim=1024,
        **kwargs,
    )



import math

import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor


__all__ = ["Transformer", "transformer_base", "transformer_big"]


class _PositionalEncoding(nn.Module):
    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000) -> None:
        super().__init__()
        self.dropout = nn.Dropout(dropout)

        pe = lucid.zeros(max_len, d_model)
        position = lucid.arange(0, max_len).unsqueeze(axis=1)
        div_term = lucid.exp(lucid.arange(0, d_model, 2) * (-math.log(1e4) / d_model))

        pe[:, 0::2] = lucid.sin(position * div_term)
        pe[:, 1::2] = lucid.cos(position * div_term)

        pe = pe.unsqueeze(axis=1)
        self.register_buffer("pe", pe)

    def forward(self, x: Tensor) -> Tensor:
        x += self.pe[: x.shape[0]]
        x = self.dropout(x)

        return x


class Transformer(nn.Module):
    def __init__(
        self,
        src_vocab_size: int,
        tgt_vocab_size: int,
        d_model: int,
        num_heads: int,
        num_encoder_layers: int,
        num_decoder_layers: int,
        dim_feedforward: int,
        dropout: float = 0.1,
    ) -> None:
        super().__init__()
        self.d_model = d_model

        self.src_tok_emb = nn.Embedding(src_vocab_size, d_model)
        self.tgt_tok_emb = nn.Embedding(tgt_vocab_size, d_model)

        self.positional_encoding = _PositionalEncoding(d_model, dropout)

        self.transformer = nn.Transformer(
            d_model=d_model,
            num_heads=num_heads,
            num_encoder_layers=num_encoder_layers,
            num_decoder_layers=num_decoder_layers,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
        )
        self.fc_out = nn.Linear(d_model, tgt_vocab_size)

    def forward(
        self,
        src: Tensor,
        tgt: Tensor,
        src_mask: Tensor | None = None,
        tgt_mask: Tensor | None = None,
        mem_mask: Tensor | None = None,
        src_key_padding_mask: Tensor | None = None,
        tgt_key_padding_mask: Tensor | None = None,
        mem_key_padding_mask: Tensor | None = None,
    ) -> Tensor:
        src_emb = self.src_tok_emb(src) * math.sqrt(self.d_model)
        tgt_emb = self.tgt_tok_emb(tgt) * math.sqrt(self.d_model)

        src_emb = self.positional_encoding(src_emb)
        tgt_emb = self.positional_encoding(tgt_emb)

        output = self.transformer(
            src_emb,
            tgt_emb,
            src_mask=src_mask,
            tgt_mask=tgt_mask,
            mem_mask=mem_mask,
            src_key_padding_mask=src_key_padding_mask,
            tgt_key_padding_mask=tgt_key_padding_mask,
            mem_key_padding_mask=mem_key_padding_mask,
        )
        output = self.fc_out(output)

        return output


@register_model
def transformer_base(
    src_vocab_size: int = 12000, tgt_vocab_size: int = 12000
) -> Transformer:
    return Transformer(
        src_vocab_size,
        tgt_vocab_size,
        d_model=512,
        num_heads=8,
        num_encoder_layers=6,
        num_decoder_layers=6,
        dim_feedforward=2048,
        dropout=0.1,
    )


@register_model
def transformer_big(
    src_vocab_size: int = 12000, tgt_vocab_size: int = 12000
) -> Transformer:
    return Transformer(
        src_vocab_size,
        tgt_vocab_size,
        d_model=1024,
        num_heads=16,
        num_encoder_layers=6,
        num_decoder_layers=6,
        dim_feedforward=4096,
        dropout=0.3,
    )



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, OrderedDict

import os, json, hashlib, urllib.request, tempfile, shutil

import lucid
import lucid.nn as nn

from lucid._tensor import Tensor


_WEIGHTS_DIR: Path = Path(__file__).resolve().parent
WEIGHTS_REGISTRY_PATH: Path = _WEIGHTS_DIR / "registry.json"

CACHE_HOME = (
    Path(os.environ.get("LUCID_HOME", Path.home() / ".cache" / "lucid")) / "weights"
)
BASE_URL = os.environ.get("LUCID_WEIGHTS_BASE", "")
MIRRORS = [
    s for s in os.environ.get("LUCID_WEIGHTS_MIRRORS", "").split(",") if s.strip()
]
OFFLINE = os.environ.get("LUCID_WEIGHTS_OFFLINE", "0") == "1"
HF_ENDPOINT = os.environ.get("HF_ENDPOINT", "https://huggingface.co")
HF_TOKEN = (
    os.environ.get("LUCID_HF_TOKEN")
    or os.environ.get("HF_TOKEN")
    or os.environ.get("HUGGINGFACEHUB_API_TOKEN")
)


@dataclass(frozen=True)
class WeightEntry:
    url: str
    sha256: str
    tag: str
    dataset: str | None = None
    meta: dict[str, Any] | None = None


def _load_json(path: Path) -> dict[str, Any]:
    if not path.exists():
        return {}
    with open(path, "r") as f:
        return json.load(f)


_REGISTRY = _load_json(WEIGHTS_REGISTRY_PATH)
_MODELS_META = _load_json(lucid.MODELS_REGISTRY_PATH)

_DYNAMIC: dict[str, Enum] = {}


def _canon(x: str) -> str:
    return x.lower()


def _family_of(model_key: str) -> str | None:
    if model_key in _MODELS_META:
        return _MODELS_META[model_key].get("family")
    return None


def _classname(model_key: str, family: str | None = None) -> str:
    if family is not None:
        return family + model_key[len(family) :].upper() + "_Weights"

    fam = _family_of(model_key)
    if fam:
        suf = model_key
        fam_key = fam.lower().replace(" ", "")
        suf = suf[len(fam_key) :] if suf.replace("_", "").startswith(fam_key) else suf
        suf = suf.lstrip()

        return f"{fam}{suf.upper()}_Weights" if suf else f"{fam}_Weights"

    parts = model_key.split("_")
    return "_".join(p.capitalize() for p in parts) + "_Weights"


def _make_enum(model_key: str, entries: dict[str, Any]) -> type[Enum]:
    members = {}
    for tag, info in entries.items():
        members[tag] = WeightEntry(
            url=info["url"],
            sha256=info["sha256"],
            tag=tag,
            dataset=info.get("dataset"),
            meta=info.get("meta", {}),
        )

    cl = Enum(
        _classname(model_key, family=entries["DEFAULT"]["meta"].get("family", None)),
        members,
    )
    _DYNAMIC[_canon(model_key)] = cl
    return cl


_created_names: list[str] = []
for k, entries in _REGISTRY.items():
    cl = _make_enum(k, entries)
    globals()[cl.__name__] = cl
    _created_names.append(cl.__name__)


def get_enum(model_key: str) -> type[Enum]:
    key = _canon(model_key)
    if key not in _DYNAMIC:
        raise KeyError(f"No weights registered for '{model_key}'.")
    return _DYNAMIC[key]


def resolve(weights: Enum | str, *, model_key: str | None = None) -> WeightEntry:
    if isinstance(weights, Enum):
        return weights.value
    if model_key is None:
        raise ValueError("model_key is required when resolving from string.")

    enum = get_enum(model_key)
    try:
        return getattr(enum, weights).value
    except AttributeError:
        return getattr(enum, "DEFAULT").value


def _hash_file(path: Path) -> str:
    h = hashlib.sha256()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(1024 * 1024), b""):
            h.update(chunk)
    return h.hexdigest()


def _hf_build_url(repo_id: str, revision: str, filename: str) -> str:
    rid = repo_id.strip("/")
    rev = revision or "main"
    return f"{HF_ENDPOINT.rstrip("/")}/{rid}/resolve/{rev}/{filename}?download=true"


def _parse_hf_uri(uri: str) -> tuple[str, str, str]:
    if not uri.startswith("hf://"):
        raise ValueError("Not an hf:// URI")

    s = uri[5:].lstrip("/")
    parts = s.split("/")
    if len(parts) < 3:
        raise ValueError("HF URI must be 'hf://owner/repo[@rev]/path'")

    owner = parts[0]
    repo_and_rev = parts[1]
    file_path = "/".join(parts[2:])

    if "@" in repo_and_rev:
        repo, rev = repo_and_rev.split("@", 1)
    else:
        repo, rev = repo_and_rev, "main"

    repo_id = f"{owner}/{repo}"
    return repo_id, rev, file_path


def _hf_candidate_urls(entry: WeightEntry) -> list[str]:
    urls: list[str] = []
    try:
        if entry.url.startswith("hf://"):
            repo_id, rev, file_path = _parse_hf_uri(entry.url)
            urls.append(_hf_build_url(repo_id, rev, file_path))
    except Exception:
        pass

    m = entry.meta or {}
    repo = m.get("hf_repo")
    file = m.get("hf_filename") or m.get("hf_path")
    rev = m.get("hf_revision", "main")

    if repo and file:
        urls.append(_hf_build_url(repo, rev, file))
    return urls


def _candidate_urls(entry: WeightEntry) -> list[str]:
    urls = []
    urls.extend(_hf_candidate_urls(entry))

    if entry.url and not entry.url.startswith("hf://"):
        urls.append(entry.url)
        if BASE_URL and entry.url.startswith("http"):
            tail = "/".join(entry.url.split("/")[3:])
            urls.insert(0, BASE_URL.rstrip("/") + "/" + tail)

        for m in MIRRORS:
            m = m.strip()
            if not m:
                continue
            if entry.url.startswith("http"):
                tail = "/".join(entry.url.split("/")[3:])
                urls.append(m.rstrip("/") + "/" + tail)
            else:
                urls.append(m.rstrip("/") + "/" + entry.url.lstrip("/"))

    return urls


def _cache_path(model_key: str, entry: WeightEntry) -> Path:
    fname = f"{_canon(model_key)}--{entry.tag}--{entry.sha256[:8]}.safetensors"
    return CACHE_HOME / fname


def load_state_dict_from_url(
    model_key: str,
    entry: WeightEntry,
    *,
    cache_dir: Path | None = None,
    force: bool = False,
) -> dict:
    cache = cache_dir or CACHE_HOME
    cache.mkdir(parents=True, exist_ok=True)
    dst = _cache_path(model_key, entry)

    if not OFFLINE and (force or not dst.exists()):
        tmp_path = Path(tempfile.mkstemp(prefix="lucid-", suffix=".tmp")[1])
        last_err = None
        try:
            for u in _candidate_urls(entry):
                try:
                    req = urllib.request.Request(u)
                    if HF_TOKEN and (HF_ENDPOINT in u or "huggingface.co" in u):
                        req.add_header("Authorization", f"Bearer {HF_TOKEN}")
                    with urllib.request.urlopen(req) as r, open(tmp_path, "wb") as out:
                        shutil.copyfileobj(r, out)
                    break

                except Exception as e:
                    last_err = e
                    continue
            else:
                if last_err:
                    raise last_err

            if _hash_file(tmp_path) != entry.sha256:
                raise RuntimeError("Checksum mismatch for downloaded weights.")
            shutil.move(tmp_path, dst)

        finally:
            try:
                tmp_path.unlink(missing_ok=True)
            except Exception:
                pass

    elif not dst.exists():
        raise FileNotFoundError(
            "Weights not present in cache and offline mode is enabled."
        )

    if _hash_file(dst) != entry.sha256:
        dst.unlink()
        return load_state_dict_from_url(model_key, entry, cache_dir=cache, force=True)

    try:
        from safetensors.numpy import load_file
    except Exception as e:
        raise ImportError(
            "safetensors is required to save .safetensors files. "
            "Install with `pip install safetensors`."
        ) from e

    state = load_file(str(dst))
    return OrderedDict((k, v) for k, v in state.items())


def apply(model: nn.Module, weights: Enum | str, *, strict: bool = True) -> Path:
    key = getattr(model, "_alt_name", type(model).__name__).lower()
    entry = resolve(weights, model_key=key)
    state = load_state_dict_from_url(key, entry)

    device = getattr(model, "device", "cpu")
    if device == "gpu":
        cooked = {}
        for k, v in state.items():
            if hasattr(v, "shape"):
                cooked[k] = Tensor(v, device="gpu")
            else:
                cooked[k] = v
        state = cooked

    model.load_state_dict(state, strict=strict)
    return _cache_path(key, entry)


def list_available(model_key: str) -> dict[str, WeightEntry]:
    key = _canon(model_key)
    if key not in _REGISTRY:
        return {}
    enum = get_enum(key)
    return {m.name: m.value for m in enum}


__all__ = [
    "WeightEntry",
    "get_enum",
    "resolve",
    "apply",
    "list_available",
    "load_state_dict_from_url",
    *_created_names,
]



from lucid import register_model

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.types import _DeviceType

from lucid.models.imgclf.cspnet import csp_darknet_53
from lucid.models.objdet.util import DetectionDict


__all__ = ["YOLO_V4", "yolo_v4"]


DEFAULT_ANCHORS: list[list[tuple[int, int]]] = [
    [(12, 16), (19, 36), (40, 28)],
    [(36, 75), (76, 55), (72, 146)],
    [(142, 110), (192, 243), (459, 401)],
]
DEFAULT_STRIDES: list[int] = [8, 16, 32]


def _to_xyxy(xywh: Tensor) -> tuple[Tensor, Tensor]:
    c = xywh[..., :2]
    wh = xywh[..., 2:]
    x1y1 = c - wh * 0.5
    x2y2 = c + wh * 0.5
    return x1y1, x2y2


def _bbox_iou_xywh(pred_xywh: Tensor, tgt_xywh: Tensor, eps: float = 1e-7) -> Tensor:
    p1, p2 = _to_xyxy(pred_xywh)
    t1, t2 = _to_xyxy(tgt_xywh)

    inter_lt = lucid.maximum(p1, t1)
    inter_rb = lucid.minimum(p2, t2)
    inter_wh = (inter_rb - inter_lt).clip(0)
    inter = inter_wh[..., 0] * inter_wh[..., 1]

    area_p = (p2[..., 0] - p1[..., 0]).clip(0) * (p2[..., 1] - p1[..., 1]).clip(0)
    area_t = (t2[..., 0] - t1[..., 0]).clip(0) * (t2[..., 1] - t1[..., 1]).clip(0)
    union = area_p + area_t - inter + eps
    return inter / union


def _bbox_iou_ciou(pred_xywh: Tensor, tgt_xywh: Tensor, eps: float = 1e-7) -> Tensor:
    p1, p2 = _to_xyxy(pred_xywh)
    t1, t2 = _to_xyxy(tgt_xywh)
    iou = _bbox_iou_xywh(pred_xywh, tgt_xywh, eps)

    enc_lt = lucid.minimum(p1, t1)
    enc_rb = lucid.maximum(p2, t2)
    enc_wh = (enc_rb - enc_lt).clip(0)
    c2 = enc_wh[..., 0] ** 2 + enc_wh[..., 1] ** 2 + eps

    rho2 = (pred_xywh[..., 0] - tgt_xywh[..., 0]) ** 2
    rho2 += (pred_xywh[..., 1] - tgt_xywh[..., 1]) ** 2

    v = (4 / lucid.pi**2) * (
        lucid.arctan(tgt_xywh[..., 2] / (tgt_xywh[..., 3] + eps))
        - lucid.arctan(pred_xywh[..., 2] / (pred_xywh[..., 3] + eps))
    ) ** 2

    with lucid.no_grad():
        alpha = v / (1 - iou + v + eps)
    return iou - (rho2 / c2) - alpha * v


def _iou_xyxy(a: Tensor, b: Tensor, eps: float = 1e-7) -> Tensor:
    area_a = (a[:, 2] - a[:, 0]).clip(0) * (a[:, 3] - a[:, 1]).clip(0)
    area_b = (b[:, 2] - b[:, 0]).clip(0) * (b[:, 3] - b[:, 1]).clip(0)

    tl = lucid.maximum(a[:, None, :2], b[:, :2])
    br = lucid.minimum(a[:, None, 2:], b[:, 2:])
    wh = (br - tl).clip(0)
    inter = wh[:, :, 0] * wh[:, :, 1]

    return inter / (area_a[:, None] + area_b - inter + eps)


def _diou_xyxy(a: Tensor, b: Tensor, eps: float = 1e-7) -> Tensor:
    iou = _iou_xyxy(a, b)
    ac = (a[:, :2] + a[:, 2:]) * 0.5
    bc = (b[:, :2] + b[:, 2:]) * 0.5

    diff = ac[:, None, :] - bc[None, :, :]
    rho2 = (diff**2).sum(axis=-1)

    enc_tl = lucid.minimum(a[:, None, :2], b[:, :2])
    enc_br = lucid.maximum(a[:, None, 2:], b[:, 2:])
    enc_wh = (enc_br - enc_tl).clip(0)

    c2 = (enc_wh**2).sum(axis=-1) + eps
    return iou - (rho2 / c2)


def _smooth_bce(eps: float = 0.0) -> tuple[float, float]:
    return 1.0 - 0.5 * eps, 0.5 * eps


class _DefaultCSPDarkNet53(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.net = csp_darknet_53(act=nn.Mish)

    def forward(self, x: Tensor) -> tuple[Tensor, Tensor, Tensor]:
        outs = self.net.forward_features(x, return_stage_out=True)
        H, W = x.shape[-2:]

        chosen: dict[int, Tensor] = {}
        for f in outs:
            fh, fw = f.shape[-2], f.shape[-1]

            sh = int(round(H / max(fh, 1)))
            sw = int(round(W / max(fw, 1)))
            s = max(sh, sw)
            if s in (8, 16, 32):
                chosen[s] = f

        return chosen[8], chosen[16], chosen[32]


class _ConvBNAct(nn.Module):
    def __init__(
        self,
        c_in: int,
        c_out: int,
        k: int = 1,
        s: int = 1,
        p: int | None = None,
        act: bool = True,
    ) -> None:
        super().__init__()
        if p is None:
            p = k // 2
        self.conv = nn.Conv2d(c_in, c_out, k, s, p, bias=False)
        self.bn = nn.BatchNorm2d(c_out)
        self.act = nn.LeakyReLU(0.1) if act else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        return self.act(self.bn(self.conv(x)))


class _SPPBlock(nn.Module):
    def __init__(self, c_in: int, c_mid: int | None = None) -> None:
        super().__init__()
        c_mid = c_mid if c_mid is not None else c_in // 2
        self.conv1 = _ConvBNAct(c_in, c_mid, k=1)
        self.conv2 = _ConvBNAct(c_mid, c_in, k=3)
        self.conv3 = _ConvBNAct(c_in, c_mid, k=1)

        self.pools = nn.ModuleList(
            [nn.MaxPool2d(kernel_size=k, stride=1, padding=k // 2) for k in (5, 9, 13)]
        )

        self.conv4 = _ConvBNAct(c_mid * 4, c_mid, k=1)
        self.conv5 = _ConvBNAct(c_mid, c_in, k=3)
        self.conv6 = _ConvBNAct(c_in, c_mid, k=1)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)

        pooled = [x] + [pool(x) for pool in self.pools]
        x = lucid.concatenate(pooled, axis=1)

        x = self.conv4(x)
        x = self.conv5(x)
        x = self.conv6(x)
        return x


class _FiveConv(nn.Module):
    def __init__(self, in_ch: int, mid_ch: int, out_ch: int) -> None:
        super().__init__()
        self.seq = nn.Sequential(
            _ConvBNAct(in_ch, mid_ch, k=1),
            _ConvBNAct(mid_ch, mid_ch * 2, k=3),
            _ConvBNAct(mid_ch * 2, mid_ch, k=1),
            _ConvBNAct(mid_ch, mid_ch * 2, k=3),
            _ConvBNAct(mid_ch * 2, out_ch, k=1),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.seq(x)


class _PANetNeck(nn.Module):
    def __init__(
        self,
        in_channels: list[int] = (256, 512, 1024),
        out_channels: list[int] = (256, 512, 1024),
    ) -> None:
        super().__init__()
        c3, c4, c5 = in_channels
        o3, o4, o5 = out_channels
        assert c3 == o3 and c4 == o4 and c5 == o5

        self.spp = _SPPBlock(c5)

        self.reduce_c5 = _ConvBNAct(c5 // 2, c4, k=1)
        self.c4_lat = _ConvBNAct(c4, c4, k=1)
        self.c4_td = _FiveConv(c4 * 2, c4, c4)

        self.reduce_c4 = _ConvBNAct(c4, c3, k=1)
        self.c3_lat = _ConvBNAct(c3, c3, k=1)
        self.c3_out = _FiveConv(c3 * 2, c3, c3)

        self.down_c3 = _ConvBNAct(o3, c4, k=3, s=2)
        self.c4_out = _FiveConv(c4 * 2, c4, o4)

        self.down_c4 = _ConvBNAct(o4, c4, k=3, s=2)
        self.c5_out = _FiveConv(c4 + c5 // 2, c5 // 2, o5)

    @staticmethod
    def _upsample(x: Tensor, ref: Tensor) -> Tensor:
        return F.interpolate(x, size=ref.shape[-2:], mode="nearest")

    @staticmethod
    def _match_hw(x: Tensor, ref: Tensor) -> Tensor:
        return (
            x
            if x.shape[-2:] == ref.shape[-2:]
            else F.interpolate(x, size=ref.shape[-2:], mode="nearest")
        )

    def forward(self, feats: tuple[Tensor, ...]) -> tuple[Tensor, ...]:
        c3, c4, c5 = feats
        p5_spp = self.spp(c5)

        u5 = self._upsample(self.reduce_c5(p5_spp), c4)
        p4_td = self.c4_td(lucid.concatenate([self.c4_lat(c4), u5], axis=1))

        u4 = self._upsample(self.reduce_c4(p4_td), c3)
        p3 = self.c3_out(lucid.concatenate([self.c3_lat(c3), u4], axis=1))

        d3 = self._match_hw(self.down_c3(p3), p4_td)
        p4 = self.c4_out(lucid.concatenate([d3, p4_td], axis=1))

        d4 = self._match_hw(self.down_c4(p4), p5_spp)
        p5 = self.c5_out(lucid.concatenate([d4, p5_spp], axis=1))

        return p3, p4, p5


class _YOLOHead(nn.Module):
    def __init__(
        self,
        in_channels: tuple[int, int, int],
        num_anchors: int,
        num_classes: int,
        iou_aware_alpha: float = 0.0,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.detect = nn.ModuleList()

        out_per_anchor = 6 + num_classes + (1 if iou_aware_alpha > 0 else 0)
        for c in in_channels:
            self.detect.append(
                nn.Sequential(
                    _ConvBNAct(c, c, k=3),
                    nn.Conv2d(c, num_anchors * out_per_anchor, kernel_size=1),
                )
            )

    def forward(self, feats: tuple[Tensor, ...]) -> tuple[Tensor, ...]:
        return tuple(self.detect[i](f) for i, f in enumerate(feats))


class YOLO_V4(nn.Module):
    def __init__(
        self,
        num_classes: int,
        anchors: list[list[tuple[int, int]]] | None = None,
        strides: list[int] | None = None,
        backbone: nn.Module | None = None,
        backbone_out_channels: tuple[int, int, int] | None = None,
        in_channels: tuple[int, int, int] = (256, 512, 1024),
        pos_iou_thr: float = 0.25,
        ignore_iou_thr: float = 0.5,
        obj_balance: tuple[float, float, float] = (4.0, 1.0, 0.4),
        cls_label_smoothing: float = 0.0,
        iou_aware_alpha: float = 0.5,
        iou_branch_weight: float = 1.0,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.anchors = anchors if anchors is not None else DEFAULT_ANCHORS
        self.strides = strides if strides is not None else DEFAULT_STRIDES

        self.backbone = backbone if backbone is not None else _DefaultCSPDarkNet53()
        self.neck = _PANetNeck(in_channels, in_channels)
        self.head = _YOLOHead(in_channels, len(self.anchors[0]), num_classes)

        if backbone is None:
            assert backbone_out_channels is None
            backbone_out_channels = (128, 256, 512)
        else:
            assert backbone_out_channels is not None

        self.backbone_out_channels = backbone_out_channels
        self.in_channels = in_channels

        self.c3_conv = _ConvBNAct(backbone_out_channels[0], in_channels[0], k=1)
        self.c4_conv = _ConvBNAct(backbone_out_channels[1], in_channels[1], k=1)
        self.c5_conv = _ConvBNAct(backbone_out_channels[2], in_channels[2], k=1)

        self.pos_iou_thr = pos_iou_thr
        self.ignore_iou_thr = ignore_iou_thr
        self.obj_balance = obj_balance
        self.cls_eps = cls_label_smoothing

        self.iou_aware_alpha = iou_aware_alpha
        self.iou_branch_weight = iou_branch_weight

        self._bce = nn.BCEWithLogitsLoss(reduction="mean")

    def forward(self, x: Tensor) -> tuple[Tensor, ...]:
        c3, c4, c5 = self.backbone(x)
        if (c3.shape[1], c4.shape[1], c5.shape[1]) != self.in_channels:
            c3, c4, c5 = (
                self.c3_conv(c3),
                self.c4_conv(c4),
                self.c5_conv(c5),
            )

        p3, p4, p5 = self.neck((c3, c4, c5))
        outs = self.head((p3, p4, p5))
        return outs

    def _decode_outputs(
        self, preds: tuple[Tensor, ...], img_size: tuple[int, int]
    ) -> tuple[Tensor, ...]:
        device = preds[0].device
        B = preds[0].shape[0]
        outputs = [[] for _ in range(B)]

        for s, p in enumerate(preds):
            A = len(self.anchors[s])
            C = self.num_classes
            B_, _, H, W = p.shape
            assert B == B_

            p = p.reshape(B, A, 6 + C, H, W).transpose((0, 1, 3, 4, 2))
            yv, xv = lucid.meshgrid(
                lucid.arange(H, device=device),
                lucid.arange(W, device=device),
                indexing="ij",
            )

            grid = lucid.stack([xv, yv], axis=2).reshape(1, 1, H, W, 2)
            anc = (
                lucid.tensor(self.anchors[s], device=device)
                .astype(lucid.Float)
                .reshape(1, A, 1, 1, 2)
            )
            stride = self.strides[s]

            xy = (F.sigmoid(p[..., 0:2]) + grid) * stride
            wh = F.exp(p[..., 2:4]) * anc
            objp = F.sigmoid(p[..., 4])
            ioup = F.sigmoid(p[..., 5])
            cl = F.sigmoid(p[..., 6:])

            alpha = self.iou_aware_alpha
            fused = (objp ** (1 - alpha)) * (ioup**alpha)

            x1y1 = xy - wh * 0.5
            x2y2 = xy + wh * 0.5
            x1y1[..., 0] = x1y1[..., 0].clip(0, img_size[1] - 1)
            x1y1[..., 1] = x1y1[..., 1].clip(0, img_size[0] - 1)
            x2y2[..., 0] = x2y2[..., 0].clip(0, img_size[1] - 1)
            x2y2[..., 1] = x2y2[..., 1].clip(0, img_size[0] - 1)

            out = lucid.concatenate([x1y1, x2y2, fused.unsqueeze(-1), cl], axis=-1)
            out = out.reshape(B, -1, 5 + C)

            for b in range(B):
                outputs[b].append(out[b])

        return tuple(lucid.concatenate(o, axis=0) for o in outputs)

    @lucid.no_grad()
    def _diou_nms_per_img(
        self, det: Tensor, conf_thresh: float, diou_thresh: float, max_det: int = 300
    ) -> Tensor:
        if det.size == 0:
            return lucid.zeros(0, 6, requires_grad=det.requires_grad, device=det.device)

        obj, cl = det[:, 4:5], det[:, 5:]
        conf, cl_id = (obj * cl).max(axis=1)
        mask = conf > conf_thresh
        if mask.sum() == 0:
            return lucid.zeros(0, 6, requires_grad=det.requires_grad, device=det.device)

        boxes = det[mask, :4]
        conf = conf[mask]
        cl_id = cl_id[mask].astype(lucid.Int)

        keep_idx = []
        for c in lucid.unique(cl_id):
            idx = lucid.nonzero(cl_id == c).squeeze(axis=1)
            b = boxes[idx]
            s = conf[idx]

            order = lucid.argsort(s, descending=True)
            idx = idx[order]
            b = b[order]

            while idx.size > 0:
                i = idx[0]
                keep_idx.append(i.item())
                if idx.size == 1:
                    break

                diou = _diou_xyxy(b[:1], b[1:]).squeeze(axis=0)
                mask_ = diou <= diou_thresh
                idx = idx[1:][mask_]
                b = b[1:][mask_]

        keep = lucid.tensor(keep_idx, device=det.device, dtype=lucid.Int32)[:max_det]
        return lucid.concatenate(
            [boxes[keep], conf[keep, None], cl_id[keep, None]], axis=1
        )

    def _build_targets(
        self,
        targets: list[Tensor],
        feat_shapes: list[tuple[int, int]],
        device: _DeviceType,
    ) -> tuple[list[Tensor], ...]:
        B = len(targets)
        tcls, tbox, tindices, tobj, ignore = [], [], [], [], []
        for s in range(3):
            na = len(self.anchors[s])
            H, W = feat_shapes[s]

            tcls.append([])
            tbox.append([])
            tindices.append([])

            tobj.append(lucid.zeros(B, na, H, W, device=device))
            ignore.append(lucid.zeros(B, na, H, W, dtype=bool, device=device))

        all_anc = lucid.tensor(
            [a for sc in self.anchors for a in sc], device=device, dtype=lucid.Float32
        )

        for b, tgt in enumerate(targets):
            if tgt.size == 0:
                continue

            cl = tgt[:, 0].astype(lucid.Int32)
            boxes = tgt[:, 1:5].astype(lucid.Float)

            gtw, gth = boxes[:, 2], boxes[:, 3]
            inter = lucid.minimum(gtw[:, None], all_anc[None, :, 0])
            inter *= lucid.minimum(gth[:, None], all_anc[None, :, 1])
            union = (
                (gtw[:, None] * gth[:, None])
                + (all_anc[None, :, 0] * all_anc[None, :, 1])
                - inter
                + 1e-9
            )
            wh_iou = inter / union

            pos_mask_all = wh_iou >= self.pos_iou_thr
            best_idx = lucid.argmax(wh_iou, axis=1, keepdims=True)
            pos_mask_all = lucid.where(
                pos_mask_all.any(axis=1, keepdims=True),
                pos_mask_all,
                F.one_hot(best_idx.squeeze(axis=1), wh_iou.shape[1], dtype=bool),
            )

            for i in range(boxes.shape[0]):
                gx, gy, gw, gh = boxes[i]
                gi_s, gj_s = [], []
                for s in range(3):
                    stride = self.strides[s]
                    gi_s.append(
                        int((gx / stride).clip(0, feat_shapes[s][1] - 1).item())
                    )
                    gj_s.append(
                        int((gy / stride).clip(0, feat_shapes[s][0] - 1).item())
                    )

                for a_global in lucid.nonzero(pos_mask_all[i]).flatten().tolist():
                    s = a_global // 3
                    a_local = a_global % 3
                    gi, gj = gi_s[s], gj_s[s]

                    tobj[s][b, a_local, gj, gi] = 1.0
                    tbox[s].append(lucid.tensor([gx, gy, gw, gh], device=device))
                    tcls[s].append(cl[i])
                    tindices[s].append((b, a_local, gj, gi))

                for a_global in range(9):
                    s = a_global // 3
                    a_local = a_global % 3
                    if (
                        wh_iou[i, a_global] >= self.ignore_iou_thr
                        and not pos_mask_all[i, a_global]
                    ):
                        gi, gj = gi_s[s], gj_s[s]
                        ignore[s][b, a_local, gj, gi] = True

        for s in range(3):
            if len(tbox[s]) == 0:
                tbox[s] = lucid.zeros(0, 4, device=device)
                tcls[s] = lucid.zeros(0, dtype=lucid.Int32, device=device)
                tindices[s] = lucid.zeros(0, 4, dtype=lucid.Int32, device=device)
            else:
                tbox[s] = lucid.stack(tbox[s], axis=0)
                tcls[s] = lucid.stack(tcls[s], axis=0)
                tindices[s] = lucid.tensor(
                    tindices[s], dtype=lucid.Int32, device=device
                )

        return tcls, tbox, tindices, tobj, ignore

    def get_loss(self, x: Tensor, targets: list[Tensor]) -> Tensor:
        B, _, H, W = x.shape
        device = x.device
        preds = self.forward(x)

        C = self.num_classes
        feat_shapes = [(p.shape[2], p.shape[3]) for p in preds]

        tcls, tbox, tindices, tobj, ignore = self._build_targets(
            targets, feat_shapes, device
        )
        pos_t, neg_t = _smooth_bce(self.cls_eps)

        lbox = lucid.zeros(1, device=device)
        lobj = lucid.zeros(1, device=device)
        lcls = lucid.zeros(1, device=device)
        liou = lucid.zeros(1, device=device)

        img_area = float(H * W)
        for s, p in enumerate(preds):
            na = len(self.anchors[s])
            Hs, Ws = feat_shapes[s]

            p = p.reshape(B, na, 6 + C, Hs, Ws).transpose((0, 1, 3, 4, 2))
            obj_logit = p[..., 4]

            pos_mask = tobj[s] > 0
            neg_mask = (~pos_mask) & (~ignore[s])

            if pos_mask.any():
                lobj_pos = self._bce(
                    obj_logit[pos_mask], lucid.ones_like(obj_logit[pos_mask])
                )
                lobj += self.obj_balance[s] * lobj_pos

            if neg_mask.any():
                lobj_neg = self._bce(
                    obj_logit[neg_mask], lucid.zeros_like(obj_logit[neg_mask])
                )
                lobj += self.obj_balance[s] * lobj_neg

            if tindices[s].shape[0] == 0:
                continue
            b, a, gj, gi = (
                tindices[s][:, 0],
                tindices[s][:, 1],
                tindices[s][:, 2],
                tindices[s][:, 3],
            )
            ps = p[b, a, gj, gi]

            stride = self.strides[s]
            anc = lucid.tensor(self.anchors[s], device=device, dtype=lucid.Float32)[a]

            px = (F.sigmoid(ps[:, 0]) + gi.astype(lucid.Float32)) * stride
            py = (F.sigmoid(ps[:, 1]) + gj.astype(lucid.Float32)) * stride
            pw = F.exp(ps[:, 2]) * anc[:, 0]
            ph = F.exp(ps[:, 3]) * anc[:, 1]

            pred_xywh = lucid.stack([px, py, pw, ph], axis=1)

            gw, gh = tbox[s][:, 2], tbox[s][:, 3]
            box_w = 2.0 - (gw * gh) / max(img_area, 1.0)
            ciou = _bbox_iou_ciou(pred_xywh, tbox[s])
            lbox += (box_w * (1.0 - ciou)).mean()

            with lucid.no_grad():
                iou_target = _bbox_iou_xywh(pred_xywh, tbox[s]).clip(0.0, 1.0)
            liou += self._bce(ps[:, 5], iou_target) * self.iou_branch_weight

            if C > 1:
                t = lucid.full((ps.shape[0], C), neg_t, device=device)
                t[lucid.arange(ps.shape[0], device=device), tcls[s]] = pos_t
                lcls += F.binary_cross_entropy_with_logits(
                    ps[:, 6:], t, reduction="mean"
                )

        return lbox + lobj + lcls + liou

    @lucid.no_grad()
    def predict(
        self, x: Tensor, conf_thresh: float = 0.25, diou_thresh: float = 0.5
    ) -> list[list[DetectionDict]]:
        self.eval()
        H, W = x.shape[2:]
        preds = self.forward(x)
        decoded = self._decode_outputs(preds, (H, W))

        results: list[list[DetectionDict]] = []
        for det in decoded:
            kept = self._diou_nms_per_img(det, conf_thresh, diou_thresh)
            objs = [
                {
                    "box": row[:4].tolist(),
                    "score": row[4].item(),
                    "class_id": int(row[5].item()),
                }
                for row in kept
            ]
            results.append(objs)

        return results


@register_model
def yolo_v4(num_classes: int = 80, **kwargs) -> YOLO_V4:
    return YOLO_V4(
        num_classes=num_classes,
        pos_iou_thr=0.213,
        ignore_iou_thr=0.7,
        obj_balance=(1.0, 1.0, 1.0),
        cls_label_smoothing=0.0,
        iou_aware_alpha=0.0,
        iou_branch_weight=0.0,
        **kwargs,
    )
